using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

public interface IRoundCallback
{
    void OnRoundStart();
    void OnRoundEnd(bool won);
}

// Controls the state of the current round (whether one is currently running or not)
// and provides update callbacks. Listens to Fighter events. If a round is won or lost,
// meaning if the last Fighter of the AntiPlayer died, play the according animation
// and SFX and end the current round.
public class RoundController : MonoBehaviour, IFighterCallback
{

    public static RoundController instance { get; private set; }

    [SerializeField] List<AudioClip> roundStartSfx;
    [SerializeField] List<AudioClip> roundEndSfx;
    [SerializeField] List<AudioClip> roundWinSfx;
    [SerializeField] List<AudioClip> roundLooseSfx;

    public List<IRoundCallback> roundCallbacks = new List<IRoundCallback>();
    public bool roundRunning { get; private set; } = false;
    public int roundCounter { get; private set; } = 1;
    // The Enemies that are currently placed on the field
    public List<Enemy> placedEnemies = new List<Enemy>();

    [SerializeField, Tooltip("After the last Fighter of a type died, how many seconds to wait until ending the current round")]
    float restartDelayAfterLastFighterDeath = 3f;

    BaseCameraController cameraController;
    int aliveEnemiesCount = 0;
    float targetRoundGameSpeed = 1f;
    bool inRoundEndAnimation = false;
    bool hasWonLastRound = false;
    float defaultFixedDeltaTime;

    public void StartRound()
    {
        if (roundRunning || inRoundEndAnimation)
        {
            return;
        }
        roundRunning = true;
        SetCurrentGameSpeed(targetRoundGameSpeed);
        roundCallbacks.ForEach(c => c.OnRoundStart());

        aliveEnemiesCount = 0;
        // Fighters clear their callbacks every round end, so we have to re-add them at round start
        GameObject.FindGameObjectWithTag("AntiPlayer").GetComponent<Fighter>().callbacks.Add(this);
        foreach (var enemy in placedEnemies)
        {
            enemy.callbacks.Add(this);
            aliveEnemiesCount++;
        }
        cameraController.PlayRandomGlobalAudioClip(roundStartSfx);
    }

    public void StopRound(bool won, bool skipSound = false)
    {
        if (!roundRunning || inRoundEndAnimation)
        {
            return;
        }
        roundRunning = false;
        SetCurrentGameSpeed(1f);
        roundCallbacks.ForEach(c => c.OnRoundEnd(won));
        if (!skipSound)
        {
            cameraController.PlayRandomGlobalAudioClip(roundEndSfx);
        }
    }
    public void ToggleRound()
    {
        if (roundRunning)
        {
            StopRound(false);
        }
        else
        {
            StartRound();
        }
    }

    public void SetCurrentGameSpeed(float multiplier)
    {
        Time.timeScale = multiplier;
    }

    void Update()
    {
        // Run FixedUpdate at normal interval, timescale-independent
        // https://docs.unity3d.com/ScriptReference/Time-timeScale.html
        // Mainly required for BaseCameraController to animate with the same speed.
        // Might also mess with other physics but it seems to be fine so far.
        Time.fixedDeltaTime = Time.timeScale * defaultFixedDeltaTime;
    }

    public void SetRoundSpeed(float multiplier)
    {
        targetRoundGameSpeed = multiplier;
        if (roundRunning && !inRoundEndAnimation)
        {
            SetCurrentGameSpeed(multiplier);
        }
    }

    void Awake()
    {
        if (instance == null)
        {
            instance = this;
        }
        else if (instance != this)
        {
            Destroy(gameObject);
        }
        DontDestroyOnLoad(gameObject);

        cameraController = Camera.main.GetComponent<BaseCameraController>();
        defaultFixedDeltaTime = Time.fixedDeltaTime;
    }

    public void OnFighterDeath(Fighter fighter)
    {
        var isAntiPlayer = fighter.CompareTag("AntiPlayer");
        if (isAntiPlayer)
        {
            roundCounter++;
            hasWonLastRound = true;
        }
        else
        {
            aliveEnemiesCount--;
        }
        if (isAntiPlayer || aliveEnemiesCount == 0)
        {
            // Slowmo-zoom to last dying Figher
            inRoundEndAnimation = true;
            var targetGameSpeed = 0.2f;
            SetCurrentGameSpeed(targetGameSpeed);
            cameraController.SetUserControlEnabled(false);
            cameraController.AnimateToPosition(fighter.transform.position + new Vector3(0f, 0.75f), 1.5f, 0.25f * targetGameSpeed);
            Invoke(nameof(StopRoundAfterDeath), targetGameSpeed * restartDelayAfterLastFighterDeath);
        }
    }

    // Invoked delayed after the last Fighter died to allow death animations to play out
    void StopRoundAfterDeath()
    {
        cameraController.PlayRandomGlobalAudioClip(hasWonLastRound ? roundWinSfx : roundLooseSfx);
        inRoundEndAnimation = false;
        StopRound(hasWonLastRound, true);
        hasWonLastRound = false;
        cameraController.SetUserControlEnabled(true);
        cameraController.AnimateToPosition(new Vector3(0, 0, 0), 5f, 0.15f);
    }

}
