using System.Collections.Generic;
using UnityEngine;

public interface IDragDropListener
{
    void OnEnemyPlaced(GameObject enemy, Opponent opponent, bool validPosition);
    void OnEnemyDeleted(GameObject enemy, Opponent opponent);
}

// Allows you to drag and drop existing Enemies around the field. Checks for valid
// drop positions; resets the dragged Enemy to its previous position of the target
// position was invalid (occluded by another Fighter). Also allows for deletion of
// active Enemies.
public class DragDestroyController : MonoBehaviour
{

    public static DragDestroyController instance { get; private set; }

    public List<IDragDropListener> listeners = new List<IDragDropListener>();

    [SerializeField] Color disabledColor;
    [SerializeField] LayerMask fighterRaycastLayer;
    [SerializeField] float fighterRaycastLength = 20f;
    [SerializeField] List<AudioClip> pickUpEnemySfx;
    [SerializeField] List<AudioClip> dropEnemySfx;
    [SerializeField] List<AudioClip> removeEnemySfx;

    bool draggingEnabled = true;
    GameObject selectedEnemy;
    SpriteRenderer selectedEnemySprite;
    Opponent selectedOpponent;
    Vector3 originalEnemyPosition;
    int originalEnemyLayer;

    int noCollisionLayer;
    BaseCameraController cameraController;

    // Sets the given Enemy to be the active Enemy that can be dragged around and places on the field.
    public void HandoffEnemy(GameObject enemy, bool skipSound = true)
    {
        if (selectedEnemy != null)
        {
            PlaceSelectedEnemy();
        }
        selectedEnemy = enemy;
        originalEnemyPosition = enemy.transform.position;
        selectedEnemySprite = enemy.GetComponent<SpriteRenderer>();
        selectedOpponent = enemy.GetComponent<Enemy>().opponentType;
        originalEnemyLayer = enemy.layer;
        // So that raycasts don't hit the current Enemy
        enemy.layer = noCollisionLayer;
        if (!skipSound)
        {
            cameraController.PlayRandomGlobalAudioClip(pickUpEnemySfx);
        }
    }

    // Sets whether Enemies can be dragged around
    public void SetDraggingEnabled(bool enabled)
    {
        draggingEnabled = enabled;
    }

    void Update()
    {
        if (!RoundController.instance.roundRunning && draggingEnabled)
        {
            if (Input.GetButtonUp("Fire2") && selectedEnemy == null)
            {
                DeleteHoveredEnemy();
            }
            if (Input.GetButtonDown("Fire1"))
            {
                SelectHoveredEnemy();
            }
            if (Input.GetButtonUp("Fire1"))
            {
                PlaceSelectedEnemy();
            }
            MoveSelectedEnemy();
        }
    }

    // Returns the Fighter at the current mouse position, optionally including the AntiPlayer
    GameObject RayCastForFighter(bool includeAntiPlayer)
    {
        Ray ray = Camera.main.ScreenPointToRay(Input.mousePosition);
        RaycastHit2D hit = Physics2D.GetRayIntersection(ray, fighterRaycastLength, fighterRaycastLayer);

        if (hit.collider == null || (!includeAntiPlayer && hit.collider.gameObject.CompareTag("AntiPlayer")))
        {
            return null;
        }

        return hit.collider.transform.gameObject;
    }

    // Deletes the Enemy under the mouse cursor
    void DeleteHoveredEnemy()
    {
        GameObject enemy = RayCastForFighter(false);

        if (enemy != null)
        {
            HandoffEnemy(enemy);
            listeners.ForEach(l => l.OnEnemyDeleted(enemy, selectedOpponent));
            Destroy(enemy);
            cameraController.PlayRandomGlobalAudioClip(removeEnemySfx);
        }
    }
    
    // Selects the Enemy currently under the mouse cursor to be the one being dragged around
    void SelectHoveredEnemy()
    {
        GameObject enemy = RayCastForFighter(false);
        if (enemy != null)
        {
            HandoffEnemy(enemy, false);
        }
    }

    // Set the current Enemie's position to the mouse position, accounting for configured texture offsets.
    // Indicates if the current position is occluded by another Fighter.
    void MoveSelectedEnemy()
    {
        if (selectedEnemy != null)
        {
            // Move Enemy to mouse position
            selectedEnemy.transform.position = GetMouseWorldPosition() - (Vector3)selectedOpponent.dragPivot;

            // Indicate that Enemy can't be placed on top of other Enemies
            selectedEnemySprite.color = RayCastForFighter(true) != null ? disabledColor : Color.white;
        }
    }

    // Drops the active Enemy at the mouse position.
    void PlaceSelectedEnemy()
    {
        if (selectedEnemy != null)
        {
            var validPosition = RayCastForFighter(true) == null;
            if (!validPosition)
            {
                // When trying to place the Enemy on another Fighter, reset it to its original position
                selectedEnemy.transform.position = originalEnemyPosition;
                selectedEnemySprite.color = Color.white;

            }
            selectedEnemy.layer = originalEnemyLayer;
            listeners.ForEach(l => l.OnEnemyPlaced(selectedEnemy, selectedOpponent, validPosition));
            selectedEnemy = null;
            cameraController.PlayRandomGlobalAudioClip(dropEnemySfx);
        }
    }

    Vector3 GetMouseWorldPosition()
    {
        Vector3 pos = Camera.main.ScreenToWorldPoint(Input.mousePosition);
        pos.z = selectedEnemy.transform.position.z;
        return pos;
    }

    void Awake()
    {
        if (instance == null)
        {
            instance = this;
        }
        else if (instance != this)
        {
            Destroy(gameObject);
        }
        DontDestroyOnLoad(gameObject);

        noCollisionLayer = LayerMask.NameToLayer("No Collision");
        cameraController = Camera.main.GetComponent<BaseCameraController>();
    }
}
