using System;
using UnityEngine;

// Fancy wrapper class for `Mathf.SmoothDamp()`. Interpolates between `currentValue`
// and `targetValue`, which may change rapidly without disturbing the interpolation.
// This entire system is completely over-engineered.
[Serializable]
public class Interpolator
{

    // Value to interpolate towards
    public float targetValue
    {
        get => _targetValue;
        set
        {
            var newValue = Mathf.Clamp(value, minValue, maxValue);
            if (!Mathf.Approximately(newValue, _targetValue) || !Mathf.Approximately(newValue, currentValue))
            {
                _targetValue = newValue;
                running = true;
            }
        }
    }
    // Whether the interpolator is still interpolating towards the target, i.e. `currentValue != targetValue`
    public bool running { get; private set; }
    // The current value of the interpolator; only updates when `Tick()` is called.
    public float currentValue;
    public float velocity;

    private float duration;
    private float minValue;
    private float maxValue;
    private float _targetValue;

    public Interpolator(float duration) : this(duration, 0) { }

    public Interpolator(float duration, float startValue) : this(duration, startValue, startValue) { }

    public Interpolator(float duration, float startValue, float targetValue) : this(duration, startValue, targetValue, float.NegativeInfinity, float.PositiveInfinity) { }

    public Interpolator(float duration, float startValue, float targetValue, float minValue, float maxValue)
    {
        this.duration = duration;
        this.minValue = minValue;
        this.maxValue = maxValue;
        currentValue = startValue;
        this.targetValue = targetValue;
    }

    // Updates and returns the value of this interpolator. Should be called once per frame.
    public float Tick()
    {
        if (running)
        {
            currentValue = Mathf.SmoothDamp(currentValue, _targetValue, ref velocity, duration, float.PositiveInfinity, Time.unscaledDeltaTime);
            if (Mathf.Abs(velocity) < 0.0025f)
            {
                running = false;
                currentValue = targetValue;
            }
        }
        return currentValue;
    }

    public void Reset()
    {
        velocity = 0;
        currentValue = 0;
        targetValue = 0;
    }

    public override string ToString() => $"Interpolator{{ Running: {running} Duration: {duration} Current Value: {currentValue} Target Value: {targetValue} Velocity: {velocity} }}";

}
